@@ -488,7 +488,7 @@ MVMObject * MVM_multi_cache_find_spesh(MVMThreadContext *tc, MVMObject *cache_ob
             : NULL;
         if (type_tuple) {
             MVMuint64 tt_offset = arg_idx >= arg_info->cs->num_pos
-                ? (arg_idx - arg_info->cs->num_pos) / 2
+                ? ((arg_idx + 1) - arg_info->cs->num_pos) / 2
                 : arg_idx;
             MVMuint32 is_rw = type_tuple[tt_offset].rw_cont;
             MVMSTable *known_type_st = NULL;
