@@ -532,12 +532,12 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
                 MVMROOT3(tc, static_frame, code_ref, outer, {
                     if (on_heap) {
                         MVMROOT(tc, frame, {
-                            MVM_spesh_log_entry(tc, id, static_frame, callsite);
+                            MVM_spesh_log_entry(tc, id, static_frame, callsite, args);
                         });
                     }
                     else {
                         MVMROOT2(tc, frame->caller, frame->static_info, {
-                            MVM_spesh_log_entry(tc, id, static_frame, callsite);
+                            MVM_spesh_log_entry(tc, id, static_frame, callsite, args);
                         });
                     }
                 });
@@ -1308,7 +1308,7 @@ MVMRegister * MVM_frame_find_lexical_by_name(MVMThreadContext *tc, MVMString *na
 
 /* Binds the specified value to the given lexical, finding it along the static
  * chain. */
-MVM_PUBLIC void MVM_frame_bind_lexical_by_name(MVMThreadContext *tc, MVMString *name, MVMuint16 type, MVMRegister *value) {
+MVM_PUBLIC void MVM_frame_bind_lexical_by_name(MVMThreadContext *tc, MVMString *name, MVMuint16 type, MVMRegister value) {
     MVMFrame *cur_frame = tc->cur_frame;
     while (cur_frame != NULL) {
         MVMLexicalRegistry *lexical_names = cur_frame->static_info->body.lexical_names;
@@ -1319,10 +1319,10 @@ MVM_PUBLIC void MVM_frame_bind_lexical_by_name(MVMThreadContext *tc, MVMString *
                 if (cur_frame->static_info->body.lexical_types[entry->value] == type) {
                     if (type == MVM_reg_obj || type == MVM_reg_str) {
                         MVM_ASSIGN_REF(tc, &(cur_frame->header),
-                            cur_frame->env[entry->value].o, value->o);
+                            cur_frame->env[entry->value].o, value.o);
                     }
                     else {
-                        cur_frame->env[entry->value] = *value;
+                        cur_frame->env[entry->value] = value;
                     }
                     return;
                 }
@@ -1623,7 +1623,7 @@ void MVM_frame_binddynlex(MVMThreadContext *tc, MVMString *name, MVMObject *valu
     if (!lex_reg) {
         char *c_name = MVM_string_utf8_encode_C_string(tc, name);
         char *waste[] = { c_name, NULL };
-        MVM_exception_throw_adhoc_free(tc, waste, "No contextual found with name '%s'",
+        MVM_exception_throw_adhoc_free(tc, waste, "Dynamic variable '%s' not found",
             c_name);
     }
     switch (type) {
