@@ -81,9 +81,13 @@ static void start_thread(void *data) {
     if (REPR(tc->thread_obj->body.invokee)->ID != MVM_REPR_ID_MVMCFunction)
         MVM_spesh_log_initialize_thread(tc, 0);
 
+    MVM_debugserver_notify_thread_creation(tc);
+
     /* Enter the interpreter, to run code. */
     MVM_interp_run(tc, thread_initial_invoke, ts);
 
+    MVM_debugserver_notify_thread_destruction(tc);
+
     /* Pop the temp root stack's ts->thread_obj, if it's still there (if we
      * cleared the temp root stack on exception at some point, it'll already be
      * gone). */
@@ -243,10 +247,11 @@ MVMint64 MVM_thread_lock_count(MVMThreadContext *tc, MVMObject *thread_obj) {
     }
 }
 
-void MVM_thread_cleanup_threads_list(MVMThreadContext *tc, MVMThread **head) {
+MVMint32 MVM_thread_cleanup_threads_list(MVMThreadContext *tc, MVMThread **head) {
     /* Assumed to be the only thread accessing the list.
      * must set next on every item. */
     MVMThread *new_list = NULL, *this = *head, *next;
+    MVMint32 alive = 0;
     *head = NULL;
     while (this) {
         next = this->body.next;
@@ -254,10 +259,11 @@ void MVM_thread_cleanup_threads_list(MVMThreadContext *tc, MVMThread **head) {
             case MVM_thread_stage_starting:
             case MVM_thread_stage_waiting:
             case MVM_thread_stage_started:
+                alive++;
             case MVM_thread_stage_exited:
             case MVM_thread_stage_clearing_nursery:
                 /* push it to the new starting list */
-                this->body.next = new_list;
+                MVM_ASSIGN_REF(tc, &(this->common.header), this->body.next, new_list);
                 new_list = this;
                 break;
             case MVM_thread_stage_destroyed:
@@ -270,6 +276,38 @@ void MVM_thread_cleanup_threads_list(MVMThreadContext *tc, MVMThread **head) {
         this = next;
     }
     *head = new_list;
+    return alive;
+}
+
+static const char * thread_stage_name(MVMThreadStages stage) {
+    switch (stage) {
+    case MVM_thread_stage_unstarted:
+        return "unstarted";
+    case MVM_thread_stage_starting:
+        return "starting";
+    case MVM_thread_stage_waiting:
+        return "waiting";
+    case MVM_thread_stage_started:
+        return "started";
+    case MVM_thread_stage_exited:
+        return "exited";
+    case MVM_thread_stage_clearing_nursery:
+        return "clearing_nursery";
+    case MVM_thread_stage_destroyed:
+        return "destroyed";
+    default:
+        return "INVALID";
+    }
+}
+
+/* use this in a debugger */
+void MVM_thread_dump(MVMThreadContext *tc) {
+    MVMThread *head = tc->instance->threads;
+    for (; head != NULL; head = head->body.next) {
+        fprintf(stderr, "thread id: %d stage=%s tc=%p\n",
+                head->body.thread_id, thread_stage_name(head->body.stage),
+                head->body.tc);
+    }
 }
 
 /* Goes through all non-app-lifetime threads and joins them. */
