@@ -37,6 +37,7 @@ const char * MVM_jit_expr_operator_name(MVMThreadContext *tc, enum MVMJitExprOpe
 
 /* macros used in the expression list templates, defined here so they
    don't overwrite other definitions */
+#define CAT3(a,b,c) a b c
 #define QUOTE(x) (x)
 #define SIZEOF_MEMBER(type, member) sizeof(((type*)0)->member)
 
@@ -50,7 +51,11 @@ struct ValueDefinition {
     MVMint32 addr;
 };
 
+static MVMint32 noop_code[] = { MVM_JIT_NOOP, 0 };
 
+static struct MVMJitExprTemplate noop_template = {
+    noop_code, "ns", 2, 0, 0
+};
 
 /* Logical negation of MVMJitExprOp flags. */
 MVMint32 MVM_jit_expr_op_negate_flag(MVMThreadContext *tc, MVMint32 op) {
@@ -115,8 +120,8 @@ static MVMint32 MVM_jit_expr_add_store(MVMThreadContext *tc, MVMJitExprTree *tre
 }
 
 static MVMint32 MVM_jit_expr_add_cast(MVMThreadContext *tc, MVMJitExprTree *tree,
-                                      MVMint32 node, MVMint32 to_size, MVMint32 from_size, MVMint32 is_signed) {
-    return MVM_jit_expr_apply_template_adhoc(tc, tree, "ns....", MVM_JIT_CAST, 1, node, to_size, from_size, is_signed);
+                                      MVMint32 cast_mode, MVMint32 node, MVMint32 to_size, MVMint32 from_size) {
+    return MVM_jit_expr_apply_template_adhoc(tc, tree, "ns....", cast_mode, 1, node, to_size, from_size);
 }
 
 static MVMint32 MVM_jit_expr_add_label(MVMThreadContext *tc, MVMJitExprTree *tree, MVMint32 label) {
@@ -326,36 +331,6 @@ void MVM_jit_expr_load_operands(MVMThreadContext *tc, MVMJitExprTree *tree, MVMS
     }
 }
 
-/* This function is to check the internal consistency of a template
- * before I apply it.  I need this because I make a lot of mistakes in
- * writing templates, and debugging is hard. */
-static void check_template(MVMThreadContext *tc, const MVMJitExprTemplate *template, MVMSpeshIns *ins) {
-#if MVM_JIT_DEBUG
-    MVMint32 i;
-    for (i = 0; i < template->len; i++) {
-        switch(template->info[i]) {
-        case 0:
-            MVM_oops(tc, "JIT: Template info shorter than template length (instruction %s)", ins->info->name);
-            break;
-        case 'l':
-            if (template->code[i] >= i || template->code[i] < 0)
-                MVM_oops(tc, "JIT: Template link out of bounds (instruction: %s)", ins->info->name);
-            break;
-        case 'i':
-            if (template->code[i] < 0 ||
-                (template->code[i] >= ins->info->num_operands &&
-                 !ins_has_single_input_output_operand(ins)))
-                MVM_oops(tc, "JIT: Operand access (%d) out of bounds (instruction: %s)", template->code[i], ins->info->name);
-            break;
-        default:
-            continue;
-        }
-    }
-    if (template->info[i])
-        MVM_oops(tc, "JIT: Template info longer than template length (instruction: %s)",
-                 ins->info->name);
-#endif
-}
 
 /* Add template to nodes, filling in operands and linking tree nodes. Return template root */
 static MVMint32 apply_template(MVMThreadContext *tc, MVMJitExprTree *tree, MVMint32 len, char *info,
@@ -436,7 +411,7 @@ static void analyze_node(MVMThreadContext *tc, MVMJitTreeTraverser *traverser,
     MVMint32        nchild = MVM_JIT_EXPR_NCHILD(tree, node);
     MVMint32        *links = MVM_JIT_EXPR_LINKS(tree, node);
     MVMint32         *args = MVM_JIT_EXPR_ARGS(tree, node);
-    MVMint32     cast_mode = MVM_JIT_CAST_NONE;
+    MVMint32     cast_mode = MVM_JIT_NOOP;
     MVMint32 node_size = 0;
     MVMint32 i;
 
@@ -457,7 +432,8 @@ static void analyze_node(MVMThreadContext *tc, MVMJitTreeTraverser *traverser,
     case MVM_JIT_LOAD:
         node_size = args[0];
         break;
-    case MVM_JIT_CAST:
+    case MVM_JIT_SCAST:
+    case MVM_JIT_UCAST:
         node_size = args[0];
         break;
     case MVM_JIT_LABEL:
@@ -471,7 +447,7 @@ static void analyze_node(MVMThreadContext *tc, MVMJitTreeTraverser *traverser,
     case MVM_JIT_ADDR:
     case MVM_JIT_IDX:
         node_size = MVM_JIT_PTR_SZ;
-        cast_mode = MVM_JIT_CAST_UNSIGNED;
+        cast_mode = MVM_JIT_UCAST;
         break;
         /* signed binary operations */
     case MVM_JIT_ADD:
@@ -486,7 +462,7 @@ static void analyze_node(MVMThreadContext *tc, MVMJitTreeTraverser *traverser,
             /* arithmetic nodes use their largest operand */
             node_size = MAX(MVM_JIT_EXPR_INFO(tree, links[0])->size,
                             MVM_JIT_EXPR_INFO(tree, links[1])->size);
-            cast_mode = MVM_JIT_CAST_SIGNED;
+            cast_mode = MVM_JIT_SCAST;
             break;
         }
        /* unsigned binary operations */
@@ -497,7 +473,7 @@ static void analyze_node(MVMThreadContext *tc, MVMJitTreeTraverser *traverser,
         {
             node_size = MAX(MVM_JIT_EXPR_INFO(tree, links[0])->size,
                             MVM_JIT_EXPR_INFO(tree, links[1])->size);
-            cast_mode = MVM_JIT_CAST_UNSIGNED;
+            cast_mode = MVM_JIT_UCAST;
             break;
         }
    case MVM_JIT_FLAGVAL:
@@ -536,7 +512,7 @@ static void analyze_node(MVMThreadContext *tc, MVMJitTreeTraverser *traverser,
     MVM_JIT_EXPR_INFO(tree, node)->size = node_size;
 
     /* Insert casts as necessary */
-    if (cast_mode != MVM_JIT_CAST_NONE) {
+    if (cast_mode != MVM_JIT_NOOP) {
         for (i = 0; i < nchild; i++) {
             MVMint32 child = tree->nodes[first_child+i];
             MVMuint8 child_size = MVM_JIT_EXPR_INFO(tree, child)->size;
@@ -545,7 +521,7 @@ static void analyze_node(MVMThreadContext *tc, MVMJitTreeTraverser *traverser,
                 MVM_JIT_EXPR_INFO(tree, child)->size = node_size;
             } else if (child_size < node_size) {
                 /* Widening casts need to be handled explicitly, shrinking casts do not */
-                MVMint32 cast = MVM_jit_expr_add_cast(tc, tree, child, node_size, child_size, cast_mode);
+                MVMint32 cast = MVM_jit_expr_add_cast(tc, tree, cast_mode, child, node_size, child_size);
 #if MVM_JIT_DEBUG
                 fprintf(stderr, "Inserting %s cast from %d to %d for operator %s (child %s)\n",
                         (cast_mode == MVM_JIT_CAST_UNSIGNED ? "unsigned" : "signed"),
@@ -595,6 +571,9 @@ static void active_values_flush(MVMThreadContext *tc, MVMJitExprTree *tree,
     }
 }
 
+static MVMint32 tree_is_empty(MVMThreadContext *tc, MVMJitExprTree *tree) {
+    return tree->nodes_num == 0;
+}
 
 MVMJitExprTree * MVM_jit_expr_tree_build(MVMThreadContext *tc, MVMJitGraph *jg, MVMSpeshIterator *iter) {
     MVMSpeshGraph *sg = jg->sg;
@@ -603,7 +582,6 @@ MVMJitExprTree * MVM_jit_expr_tree_build(MVMThreadContext *tc, MVMJitGraph *jg,
     MVMJitExprTree *tree;
     MVMint32 operands[MVM_MAX_OPERANDS];
     struct ValueDefinition *values;
-    MVMint32 root, node;
     MVMuint16 i;
     /* No instructions, just skip */
     if (!iter->ins)
@@ -615,9 +593,7 @@ MVMJitExprTree * MVM_jit_expr_tree_build(MVMThreadContext *tc, MVMJitGraph *jg,
     MVM_VECTOR_INIT(tree->constants, 16);
     MVM_VECTOR_INIT(tree->roots, 16);
 
-    /* start with a no-op so every valid reference is nonzero */
-    MVM_VECTOR_PUSH(tree->nodes, MVM_JIT_NOOP);
-    MVM_VECTOR_PUSH(tree->nodes, 0);
+
 
     tree->graph      = jg;
     /* Hold indices to the node that last computed a value belonging
@@ -626,7 +602,8 @@ MVMJitExprTree * MVM_jit_expr_tree_build(MVMThreadContext *tc, MVMJitGraph *jg,
     values = MVM_malloc(sizeof(struct ValueDefinition)*sg->num_locals);
     memset(values, -1, sizeof(struct ValueDefinition)*sg->num_locals);
 
-#define BAIL(x, ...) do { if (x) { MVM_jit_log(tc, __VA_ARGS__); goto done; } } while (0)
+#define BAIL(x, ...) do { if (x) { MVM_spesh_graph_add_comment(tc, iter->graph, iter->ins, "expr bail: " __VA_ARGS__); goto done; } } while (0)
+
 
     /* Generate a tree based on templates. The basic idea is to keep a
        index to the node that last computed the value of a local.
@@ -642,13 +619,13 @@ MVMJitExprTree * MVM_jit_expr_tree_build(MVMThreadContext *tc, MVMJitGraph *jg,
         MVMuint16 opcode = ins->info->opcode;
         MVMSpeshAnn *ann;
         const MVMJitExprTemplate *template;
-        MVMint32 before_label = -1, after_label = -1, store_directly = 0;
+        MVMint32 before_label = -1, after_label = -1, store_directly = 0, root = 0;
 
         struct ValueDefinition *defined_value = NULL;
 
         /* check if this is a getlex and if we can handle it */
-        BAIL(opcode == MVM_OP_getlex && getlex_needs_autoviv(tc, jg, ins), "Can't compile object getlex\n");
-        BAIL(opcode == MVM_OP_bindlex && bindlex_needs_write_barrier(tc, jg, ins), "Can't compile write-barrier bindlex\n");
+        BAIL(opcode == MVM_OP_getlex && getlex_needs_autoviv(tc, jg, ins), "getlex with autoviv");
+        BAIL(opcode == MVM_OP_bindlex && bindlex_needs_write_barrier(tc, jg, ins), "Can't compile write-barrier bindlex");
 
         /* Check annotations that may require handling or wrapping the expression */
         for (ann = ins->annotations; ann != NULL; ann = ann->next) {
@@ -701,7 +678,6 @@ MVMJitExprTree * MVM_jit_expr_tree_build(MVMThreadContext *tc, MVMJitGraph *jg,
                 jg->inlines[ann->data.inline_idx].end_label = after_label;
                 break;
             case MVM_SPESH_ANN_DEOPT_INLINE:
-                MVM_jit_log(tc, "Not sure if we can handle DEOPT_INLINE on instruction %s\n", ins->info->name);
                 break;
             case MVM_SPESH_ANN_DEOPT_ONE_INS:
                 /* we should only see this in guards, which we don't do just
@@ -711,7 +687,7 @@ MVMJitExprTree * MVM_jit_expr_tree_build(MVMThreadContext *tc, MVMJitGraph *jg,
                 case MVM_OP_sp_guardconc:
                 case MVM_OP_sp_guardtype:
                 case MVM_OP_sp_guardsf:
-                    BAIL(1, "Cannot handle DEOPT_ONE (ins=%s)\n", ins->info->name);
+                    BAIL(1, "Cannot handle DEOPT_ONE (ins=%s)", ins->info->name);
                     break;
                 }
                 break;
@@ -731,28 +707,20 @@ MVMJitExprTree * MVM_jit_expr_tree_build(MVMThreadContext *tc, MVMJitGraph *jg,
 
 
         if (opcode == MVM_SSA_PHI || opcode == MVM_OP_no_op) {
-            /* By definition, a PHI node can only occur at the start of a basic
-             * block. (A no_op instruction only seems to happen as the very
-             * first instruction of a frame, and I'm not sure why).
-             *
-             * Thus, if it happens that we've processed annotations on those
-             * instructions (which probably means they migrated there from
-             * somewhere else), they always refer to the start of the basic
-             * block, which is already assigned a label and
-             * dynamic-control-handler.
-             *
-             * So we never need to do anything with this label and wrapper, but
-             * we do need to process the annotation to setup the frame handler
-             * correctly.
-             */
-            BAIL(after_label >= 0, "A PHI node should not have an after label\n");
-            continue;
+            /* No template here, but we may have to emit labels */
+            if (after_label < 0 && (before_label < 0 || tree_is_empty(tc, tree)))
+                continue;
+            goto emit;
         }
 
         template = MVM_jit_get_template_for_opcode(opcode);
-        BAIL(template == NULL, "Cannot get template for: %s\n", ins->info->name);
+        BAIL(template == NULL, "Cannot get template for: %s", ins->info->name);
+        if (tree_is_empty(tc, tree)) {
+            /* start with a no-op so every valid reference is nonzero */
+            MVM_jit_expr_apply_template(tc, tree, &noop_template, NULL);
+            MVM_spesh_graph_add_comment(tc, jg->sg, iter->ins, "start of exprjit tree");
+        }
 
-        check_template(tc, template, ins);
 
         MVM_jit_expr_load_operands(tc, tree, ins, values, operands);
         root = MVM_jit_expr_apply_template(tc, tree, template, operands);
@@ -788,7 +756,7 @@ MVMJitExprTree * MVM_jit_expr_tree_build(MVMThreadContext *tc, MVMJitGraph *jg,
                     memset(values + opr.reg.orig, -1, sizeof(struct ValueDefinition));
                 } else {
                     /* record this value, should be only one for the root */
-                    BAIL(i != 0, "Write reg operand %d\n", i);
+                    BAIL(i != 0, "Write reg operand %d", i);
                     MVM_JIT_EXPR_INFO(tree, root)->type = opr_type >> 3;
                     defined_value = values + opr.reg.orig;
                     defined_value->addr = operands[i];
@@ -801,7 +769,7 @@ MVMJitExprTree * MVM_jit_expr_tree_build(MVMThreadContext *tc, MVMJitGraph *jg,
                 /* does not define a value we can look up, but we may need to
                  * insert a store */
                 if (!(template->flags & MVM_JIT_EXPR_TEMPLATE_DESTRUCTIVE)) {
-                    BAIL(i != 0, "Write lex operand %d\n", i);
+                    BAIL(i != 0, "Write lex operand %d", i);
                     MVM_JIT_EXPR_INFO(tree, root)->type = opr_type >> 3;
                     /* insert the store to lexicals directly, do not record as value */
                     root = MVM_jit_expr_add_store(tc, tree, operands[i], root, MVM_JIT_REG_SZ);
@@ -818,7 +786,6 @@ MVMJitExprTree * MVM_jit_expr_tree_build(MVMThreadContext *tc, MVMJitGraph *jg,
              * to replace an invokish version with a non-invokish version (but
              * perhaps best if that is opt-in so people don't accidentally
              * forget to set it). */
-            MVM_jit_log(tc, "EXPR: adding throwish guard to op (%s)\n", ins->info->name);
             active_values_flush(tc, tree, values, sg->num_locals);
             store_directly = 1;
         }
@@ -826,6 +793,7 @@ MVMJitExprTree * MVM_jit_expr_tree_build(MVMThreadContext *tc, MVMJitGraph *jg,
 
         /* Add root to tree to ensure source evaluation order, wrapped with
          * labels if necessary. */
+    emit:
         if (before_label >= 0 && MVM_jit_label_is_for_ins(tc, jg, before_label)) {
             MVM_VECTOR_PUSH(tree->roots, MVM_jit_expr_add_label(tc, tree, before_label));
         }
@@ -844,7 +812,9 @@ MVMJitExprTree * MVM_jit_expr_tree_build(MVMThreadContext *tc, MVMJitGraph *jg,
             defined_value->root = tree->roots_num;
         }
 
-        MVM_VECTOR_PUSH(tree->roots, root);
+        if (root != 0)
+            MVM_VECTOR_PUSH(tree->roots, root);
+
         if (after_label >= 0 && MVM_jit_label_is_for_ins(tc, jg, after_label)) {
             MVM_VECTOR_PUSH(tree->roots, MVM_jit_expr_add_label(tc, tree, after_label));
         }
@@ -854,13 +824,6 @@ MVMJitExprTree * MVM_jit_expr_tree_build(MVMThreadContext *tc, MVMJitGraph *jg,
     if (tree->roots_num > 0) {
         active_values_flush(tc, tree, values, sg->num_locals);
         MVM_jit_expr_tree_analyze(tc, tree);
-        MVM_jit_log(tc, "Build tree out of: [");
-        for (ins = entry; ins != iter->ins; ins = ins->next) {
-            if (ins->info->opcode == MVM_SSA_PHI)
-                continue;
-            MVM_jit_log(tc, "%s, ", ins->info->name);
-        }
-        MVM_jit_log(tc, "]\n");
     } else {
         /* Don't return empty trees, nobody wants that */
         MVM_jit_expr_tree_destroy(tc, tree);
