@@ -53,12 +53,7 @@ static FILE *fopen_perhaps_with_pid(char *env_var, char *path, const char *mode)
             result = fopen(path, mode);
         } else {
             char *fixed_path = malloc(path_length + 16);
-            MVMint64 pid;
-#ifdef _WIN32
-            pid = _getpid();
-#else
-            pid = getpid();
-#endif
+            MVMint64 pid = MVM_proc_getpid(NULL);
             /* We make the brave assumption that
              * pids only go up to 16 characters. */
             snprintf(fixed_path, path_length + 16, path, pid);
@@ -82,7 +77,7 @@ MVMInstance * MVM_vm_create_instance(void) {
 
     char *spesh_log, *spesh_nodelay, *spesh_disable, *spesh_inline_disable,
          *spesh_osr_disable, *spesh_limit, *spesh_blocking, *spesh_inline_log;
-    char *jit_log, *jit_expr_disable, *jit_disable, *jit_bytecode_dir, *jit_last_frame, *jit_last_bb;
+    char *jit_expr_disable, *jit_disable, *jit_last_frame, *jit_last_bb;
     char *dynvar_log;
     int init_stat;
 
@@ -168,7 +163,7 @@ MVMInstance * MVM_vm_create_instance(void) {
     instance->main_thread->last_payload = instance->VMNull;
 
     /* Initialize event loop thread starting mutex. */
-    init_mutex(instance->mutex_event_loop_start, "event loop thread start");
+    init_mutex(instance->mutex_event_loop, "event loop thread start");
 
     /* Create main thread object, and also make it the start of the all threads
      * linked list. Set up the mutex to protect it. */
@@ -263,18 +258,43 @@ MVMInstance * MVM_vm_create_instance(void) {
         instance->jit_expr_enabled = 1;
 
 
-    jit_log = getenv("MVM_JIT_LOG");
-    if (jit_log && jit_log[0])
-        instance->jit_log_fh = fopen_perhaps_with_pid("MVM_JIT_LOG", jit_log, "w");
-    jit_bytecode_dir = getenv("MVM_JIT_BYTECODE_DIR");
-    if (jit_bytecode_dir && jit_bytecode_dir[0]) {
-        size_t bytecode_map_name_size = strlen(jit_bytecode_dir) + strlen("/jit-map.txt") + 1;
-        char *bytecode_map_name = MVM_malloc(bytecode_map_name_size);
-        snprintf(bytecode_map_name, bytecode_map_name_size, "%s/jit-map.txt", jit_bytecode_dir);
-        instance->jit_bytecode_map = fopen(bytecode_map_name, "w");
-        instance->jit_bytecode_dir = jit_bytecode_dir;
-        MVM_free(bytecode_map_name);
+    {
+        char *jit_debug = getenv("MVM_JIT_DEBUG");
+        if (jit_debug && jit_debug[0])
+            instance->jit_debug_enabled = 1;
     }
+
+#if linux
+    {
+        char *jit_perf_map = getenv("MVM_JIT_PERF_MAP");
+        if (jit_perf_map && *jit_perf_map) {
+            char perf_map_filename[32];
+            snprintf(perf_map_filename, sizeof(perf_map_filename),
+                     "/tmp/perf-%"PRIi64".map", MVM_proc_getpid(NULL));
+            instance->jit_perf_map = fopen(perf_map_filename, "w");
+        }
+    }
+#endif
+
+    {
+        char *jit_dump_bytecode = getenv("MVM_JIT_DUMP_BYTECODE");
+        if (jit_dump_bytecode && jit_dump_bytecode[0]) {
+            char tmpdir[1024];
+            size_t len = sizeof tmpdir;
+            char *jit_bytecode_dir;
+            uv_fs_t req;
+            uv_os_tmpdir(tmpdir, &len);
+            jit_bytecode_dir = MVM_malloc(len + 32);
+            snprintf(jit_bytecode_dir, len+32, "%s/moar-jit.%"PRIi64,
+                     tmpdir, MVM_proc_getpid(NULL));
+            if (uv_fs_mkdir(NULL, &req, jit_bytecode_dir, 0755, NULL) == 0) {
+                instance->jit_bytecode_dir = jit_bytecode_dir;
+            } else {
+                MVM_free(jit_bytecode_dir);
+            }
+        }
+    }
+
     jit_last_frame = getenv("MVM_JIT_EXPR_LAST_FRAME");
     jit_last_bb    = getenv("MVM_JIT_EXPR_LAST_BB");
 
@@ -362,7 +382,7 @@ MVMInstance * MVM_vm_create_instance(void) {
     setup_std_handles(instance->main_thread);
 
     /* Set up the specialization worker thread and a log for the main thread. */
-    MVM_spesh_worker_setup(instance->main_thread);
+    MVM_spesh_worker_start(instance->main_thread);
     MVM_spesh_log_initialize_thread(instance->main_thread, 1);
 
     /* Back to nursery allocation, now we're set up. */
@@ -457,10 +477,6 @@ void MVM_vm_exit(MVMInstance *instance) {
     /* Close any spesh or jit log. */
     if (instance->spesh_log_fh)
         fclose(instance->spesh_log_fh);
-    if (instance->jit_log_fh)
-        fclose(instance->jit_log_fh);
-    if (instance->jit_bytecode_map)
-        fclose(instance->jit_bytecode_map);
     if (instance->dynvar_log_fh) {
         fprintf(instance->dynvar_log_fh, "- x 0 0 0 0 %"PRId64" %"PRIu64" %"PRIu64"\n", instance->dynvar_log_lasttime, uv_hrtime(), uv_hrtime());
         fclose(instance->dynvar_log_fh);
@@ -500,10 +516,16 @@ static void cleanup_callsite_interns(MVMInstance *instance) {
  * should clear up all resources and free all memory; in practice, it falls
  * short of this goal at the moment. */
 void MVM_vm_destroy_instance(MVMInstance *instance) {
+
     /* Join any foreground threads and flush standard handles. */
     MVM_thread_join_foreground(instance->main_thread);
     MVM_io_flush_standard_handles(instance->main_thread);
 
+    /* Stop system threads */
+    MVM_spesh_worker_stop(instance->main_thread);
+    MVM_spesh_worker_join(instance->main_thread);
+    MVM_io_eventloop_destroy(instance->main_thread);
+
     /* Run the GC global destruction phase. After this,
      * no 6model object pointers should be accessed. */
     MVM_gc_global_destruction(instance->main_thread);
@@ -582,10 +604,12 @@ void MVM_vm_destroy_instance(MVMInstance *instance) {
     uv_mutex_destroy(&instance->mutex_spesh_sync);
     if (instance->spesh_log_fh)
         fclose(instance->spesh_log_fh);
-    if (instance->jit_log_fh)
-        fclose(instance->jit_log_fh);
+    if (instance->jit_perf_map)
+        fclose(instance->jit_perf_map);
     if (instance->dynvar_log_fh)
         fclose(instance->dynvar_log_fh);
+    if (instance->jit_bytecode_dir)
+        MVM_free(instance->jit_bytecode_dir);
     if (instance->jit_breakpoints) {
         MVM_VECTOR_DESTROY(instance->jit_breakpoints);
     }
@@ -598,21 +622,22 @@ void MVM_vm_destroy_instance(MVMInstance *instance) {
     uv_mutex_destroy(&instance->nfg->update_mutex);
     MVM_nfg_destroy(instance->main_thread);
 
-    /* Clean up fixed size allocator */
-    MVM_fixed_size_destroy(instance->fsa);
 
     /* Clean up integer constant and string cache. */
     uv_mutex_destroy(&instance->mutex_int_const_cache);
     MVM_free(instance->int_const_cache);
     MVM_free(instance->int_to_str_cache);
 
-    /* Clean up event loop starting mutex. */
-    uv_mutex_destroy(&instance->mutex_event_loop_start);
+    /* Clean up event loop mutex. */
+    uv_mutex_destroy(&instance->mutex_event_loop);
 
     /* Destroy main thread contexts and thread list mutex. */
     MVM_tc_destroy(instance->main_thread);
     uv_mutex_destroy(&instance->mutex_threads);
 
+    /* Clean up fixed size allocator */
+    MVM_fixed_size_destroy(instance->fsa);
+
     /* Clear up VM instance memory. */
     MVM_free(instance);
 }
