@@ -38,11 +38,12 @@ void MVM_spesh_candidate_add(MVMThreadContext *tc, MVMSpeshPlanned *p) {
     MVMSpeshCandidate *candidate;
     MVMSpeshCandidate **new_candidate_list;
     MVMStaticFrameSpesh *spesh;
-    MVMuint64 start_time;
+    MVMuint64 start_time, spesh_time, jit_time, end_time;
 
     /* If we've reached our specialization limit, don't continue. */
+    MVMint32 spesh_produced = ++tc->instance->spesh_produced;
     if (tc->instance->spesh_limit)
-        if (++tc->instance->spesh_produced > tc->instance->spesh_limit)
+        if (spesh_produced > tc->instance->spesh_limit)
             return;
 
     /* Produce the specialization graph and, if we're logging, dump it out
@@ -50,8 +51,6 @@ void MVM_spesh_candidate_add(MVMThreadContext *tc, MVMSpeshPlanned *p) {
 #if MVM_GC_DEBUG
     tc->in_spesh = 1;
 #endif
-    if (MVM_spesh_debug_enabled(tc))
-        start_time = uv_hrtime();
     sg = MVM_spesh_graph_create(tc, p->sf, 0, 1);
     if (MVM_spesh_debug_enabled(tc)) {
         char *c_name = MVM_string_utf8_encode_C_string(tc, p->sf->body.name);
@@ -64,6 +63,7 @@ void MVM_spesh_candidate_add(MVMThreadContext *tc, MVMSpeshPlanned *p) {
         MVM_free(c_cuid);
         MVM_free(before);
         fflush(tc->instance->spesh_log_fh);
+        start_time = uv_hrtime();
     }
 
     /* Attach the graph so we will be able to mark it during optimization,
@@ -84,14 +84,8 @@ void MVM_spesh_candidate_add(MVMThreadContext *tc, MVMSpeshPlanned *p) {
     /* Clear active graph; beyond this point, no more GC syncs. */
     tc->spesh_active_graph = NULL;
 
-    if (MVM_spesh_debug_enabled(tc)) {
-        char *after = MVM_spesh_dump(tc, sg);
-        MVM_spesh_debug_printf(tc, "After:\n%s", after);
-        MVM_spesh_debug_printf(tc, "Specialization took %dus\n\n========\n\n",
-            (int)((uv_hrtime() - start_time) / 1000));
-        MVM_free(after);
-        fflush(tc->instance->spesh_log_fh);
-    }
+    if (MVM_spesh_debug_enabled(tc))
+        spesh_time = uv_hrtime();
 
     /* Generate code and install it into the candidate. */
     sc = MVM_spesh_codegen(tc, sg);
@@ -116,18 +110,37 @@ void MVM_spesh_candidate_add(MVMThreadContext *tc, MVMSpeshPlanned *p) {
      * the spesh graph and can safely be deleted with it. */
     if (tc->instance->jit_enabled) {
         MVMJitGraph *jg;
-        if (tc->instance->spesh_log_fh)
-            start_time = uv_hrtime();
+        if (MVM_spesh_debug_enabled(tc))
+            jit_time = uv_hrtime();
+
         jg = MVM_jit_try_make_graph(tc, sg);
         if (jg != NULL) {
             candidate->jitcode = MVM_jit_compile_graph(tc, jg);
             MVM_jit_graph_destroy(tc, jg);
         }
-        if (MVM_spesh_debug_enabled(tc)) {
-            MVM_spesh_debug_printf(tc, "JIT was %s and compilation took %" PRIu64 "us\n",
-                                   candidate->jitcode ? "successful" : "not successful",
-                                   (uv_hrtime() - start_time) / 1000);
+    }
+
+    if (MVM_spesh_debug_enabled(tc)) {
+        char *after = MVM_spesh_dump(tc, sg);
+        end_time = uv_hrtime();
+        MVM_spesh_debug_printf(tc, "After:\n%s", after);
+        MVM_spesh_debug_printf(tc,
+            "Specialization took %" PRIu64 "us (total %" PRIu64"us)\n",
+            (spesh_time - start_time) / 1000,
+            (end_time - start_time) / 1000);
+
+        if (tc->instance->jit_enabled) {
+            MVM_spesh_debug_printf(tc,
+                "JIT was %ssuccessful and compilation took %" PRIu64 "us\n",
+                candidate->jitcode ? "" : "not ", (end_time - jit_time) / 1000);
+            if (candidate->jitcode) {
+                MVM_spesh_debug_printf(tc, "    Bytecode size: %" PRIu64 " byte\n",
+                                       candidate->jitcode->size);
+            }
         }
+        MVM_spesh_debug_printf(tc, "\n========\n\n");
+        MVM_free(after);
+        fflush(tc->instance->spesh_log_fh);
     }
 
     /* calculate work environment taking JIT spill area into account */
@@ -137,15 +150,8 @@ void MVM_spesh_candidate_add(MVMThreadContext *tc, MVMSpeshPlanned *p) {
     candidate->num_spesh_slots = sg->num_spesh_slots;
     candidate->spesh_slots     = sg->spesh_slots;
 
-    /* Clean up after specialization work. */
-    if (candidate->num_inlines) {
-        MVMint32 i;
-        for (i = 0; i < candidate->num_inlines; i++)
-            if (candidate->inlines[i].g) {
-                MVM_spesh_graph_destroy(tc, candidate->inlines[i].g);
-                candidate->inlines[i].g = NULL;
-            }
-    }
+    /* Claim ownership of allocated memory assigned to the candidate */
+    sg->cand = candidate;
     MVM_spesh_graph_destroy(tc, sg);
 
     /* Create a new candidate list and copy any existing ones. Free memory
@@ -203,4 +209,5 @@ void MVM_spesh_candidate_destroy(MVMThreadContext *tc, MVMSpeshCandidate *candid
     MVM_free(candidate->lexical_types);
     if (candidate->jitcode)
         MVM_jit_code_destroy(tc, candidate->jitcode);
+    MVM_free(candidate);
 }
