@@ -13,17 +13,16 @@ static void copy_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint16 to_orig
     tfacts->type          = ffacts->type;
     tfacts->decont_type   = ffacts->decont_type;
     tfacts->value         = ffacts->value;
-    tfacts->log_guard     = ffacts->log_guard;
+    tfacts->log_guards    = ffacts->log_guards;
+    tfacts->num_log_guards = ffacts->num_log_guards;
 }
 
 /* Called when one set of facts depend on another, allowing any log guard
  * that is to thank to be marked used as needed later on. */
 void MVM_spesh_facts_depend(MVMThreadContext *tc, MVMSpeshGraph *g,
                             MVMSpeshFacts *target, MVMSpeshFacts *source) {
-    if (source->flags & MVM_SPESH_FACT_FROM_LOG_GUARD) {
-        target->flags     |= MVM_SPESH_FACT_FROM_LOG_GUARD;
-        target->log_guard  = source->log_guard;
-    }
+    target->log_guards = source->log_guards;
+    target->num_log_guards = source->num_log_guards;
 }
 
 /* Handles object-creating instructions. */
@@ -41,14 +40,6 @@ static void create_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint16 obj_o
 
     /* We know it's a concrete object. */
     obj_facts->flags |= MVM_SPESH_FACT_CONCRETE;
-
-    /* If we know the type object, then we can check to see if
-     * it's a container type. */
-    if (type_facts->flags & MVM_SPESH_FACT_KNOWN_TYPE) {
-        MVMObject *type = type_facts->type;
-        if (type && !STABLE(type)->container_spec)
-            obj_facts->flags |= MVM_SPESH_FACT_DECONTED;
-    }
 }
 
 static void create_facts_with_type(MVMThreadContext *tc, MVMSpeshGraph *g,
@@ -62,11 +53,6 @@ static void create_facts_with_type(MVMThreadContext *tc, MVMSpeshGraph *g,
 
     /* We know it's a concrete object. */
     obj_facts->flags |= MVM_SPESH_FACT_CONCRETE;
-
-    /* If we know the type object, then we can check to see if
-     * it's a container type. */
-    if (type && !STABLE(type)->container_spec)
-        obj_facts->flags |= MVM_SPESH_FACT_DECONTED;
 }
 
 /* Adds facts from knowing the exact value being put into an object local. */
@@ -84,15 +70,11 @@ static void object_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint16 tgt_o
     g->facts[tgt_orig][tgt_i].type   = STABLE(obj)->WHAT;
     g->facts[tgt_orig][tgt_i].flags |= MVM_SPESH_FACT_KNOWN_TYPE;
 
-    /* Set concreteness and decontainerized flags. */
-    if (IS_CONCRETE(obj)) {
+    /* Set concreteness flags. */
+    if (IS_CONCRETE(obj))
         g->facts[tgt_orig][tgt_i].flags |= MVM_SPESH_FACT_CONCRETE;
-        if (!STABLE(obj)->container_spec)
-            g->facts[tgt_orig][tgt_i].flags |= MVM_SPESH_FACT_DECONTED;
-    }
-    else {
-        g->facts[tgt_orig][tgt_i].flags |= MVM_SPESH_FACT_TYPEOBJ | MVM_SPESH_FACT_DECONTED;
-    }
+    else
+        g->facts[tgt_orig][tgt_i].flags |= MVM_SPESH_FACT_TYPEOBJ;
 }
 void MVM_spesh_facts_object_facts(MVMThreadContext *tc, MVMSpeshGraph *g,
                                   MVMSpeshOperand tgt, MVMObject *obj) {
@@ -109,11 +91,12 @@ static void decont_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns *in
     /* If we know the original is decontainerized already, just copy its
      * info. */
     MVMint32 in_flags = in_facts->flags;
-    if (in_flags & MVM_SPESH_FACT_DECONTED)
+    if ((in_flags & MVM_SPESH_FACT_TYPEOBJ) ||
+            (in_flags & MVM_SPESH_FACT_KNOWN_TYPE) &&
+            !in_facts->type->st->container_spec) {
         copy_facts(tc, g, out_orig, out_i, in_orig, in_i);
-
-    /* We know the result is decontainerized. */
-    out_facts->flags |= MVM_SPESH_FACT_DECONTED;
+        return;
+    }
 
     /* We may also know the original was containerized, and have some facts
      * about its contents. */
@@ -219,6 +202,14 @@ static void getstringfrom_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpesh
     tgt_facts->value.s = str;
     tgt_facts->flags |= MVM_SPESH_FACT_KNOWN_VALUE;
 }
+static void trunc_i16_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns *ins) {
+    MVMSpeshFacts *src_facts = &g->facts[ins->operands[1].reg.orig][ins->operands[1].reg.i];
+    if (src_facts->flags & MVM_SPESH_FACT_KNOWN_VALUE) {
+        MVMSpeshFacts *tgt_facts = &g->facts[ins->operands[0].reg.orig][ins->operands[0].reg.i];
+        tgt_facts->value.i = (MVMint16)src_facts->value.i;
+        tgt_facts->flags |= MVM_SPESH_FACT_KNOWN_VALUE;
+    }
+}
 
 /* Discover facts from extops. */
 static void discover_extop(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns *ins) {
@@ -349,25 +340,23 @@ static void log_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
         MVMSpeshOperand guard_reg = ins->operands[0];
         MVMSpeshOperand preguard_reg = MVM_spesh_manipulate_new_version(tc, g,
                 ins->operands[0].reg.orig);
+        MVMSpeshFacts *pre_facts = &g->facts[preguard_reg.reg.orig][preguard_reg.reg.i];
         MVMSpeshFacts *facts = &g->facts[guard_reg.reg.orig][guard_reg.reg.i];
         ins->operands[0] = preguard_reg;
+        pre_facts->writer = ins;
 
         /* Add facts and choose guard op. */
         facts->type = agg_type;
         facts->flags |= MVM_SPESH_FACT_KNOWN_TYPE;
         if (agg_concrete && !agg_type_object) {
             facts->flags |= MVM_SPESH_FACT_CONCRETE;
-            if (!agg_type->st->container_spec)
-                facts->flags |= MVM_SPESH_FACT_DECONTED;
             guard_op = MVM_OP_sp_guardconc;
         }
         else if (agg_type_object && !agg_concrete) {
-            facts->flags |= MVM_SPESH_FACT_TYPEOBJ | MVM_SPESH_FACT_DECONTED;
+            facts->flags |= MVM_SPESH_FACT_TYPEOBJ;
             guard_op = MVM_OP_sp_guardtype;
         }
         else {
-            if (!agg_type->st->container_spec)
-                facts->flags |= MVM_SPESH_FACT_DECONTED;
             guard_op = MVM_OP_sp_guard;
         }
 
@@ -384,6 +373,8 @@ static void log_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
             MVM_spesh_manipulate_insert_ins(tc, bb, ins, guard);
         else
             MVM_spesh_manipulate_insert_ins(tc, bb->linear_next, NULL, guard);
+        facts->writer = guard;
+        MVM_spesh_usages_add_by_reg(tc, g, preguard_reg, guard);
 
         /* Move deopt annotation to the guard instruction. */
         ann = ins->annotations;
@@ -402,6 +393,10 @@ static void log_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
         deopt_one_ann->next = NULL;
         guard->annotations = deopt_one_ann;
 
+        /* Move deopt usages to the preguard register. */
+        pre_facts->usage.deopt_users = facts->usage.deopt_users;
+        facts->usage.deopt_users = NULL;
+
         /* Add entry in log guards table, and mark facts as depending on it. */
         if (g->num_log_guards % 16 == 0) {
             MVMSpeshLogGuard *orig_log_guards = g->log_guards;
@@ -413,8 +408,9 @@ static void log_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
         }
         g->log_guards[g->num_log_guards].ins = guard;
         g->log_guards[g->num_log_guards].bb = ins->next ? bb : bb->linear_next;
-        facts->flags |= MVM_SPESH_FACT_FROM_LOG_GUARD;
-        facts->log_guard = g->num_log_guards;
+        facts->log_guards = MVM_spesh_alloc(tc, g, sizeof(MVMint32));
+        facts->log_guards[0] = g->num_log_guards;
+        facts->num_log_guards++;
         g->num_log_guards++;
     }
 }
@@ -445,37 +441,8 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
         if (p && ann_deopt_one && ann_logged && ins->info->opcode != MVM_OP_speshresolve)
             log_facts(tc, g, bb, ins, p, ann_deopt_one, ann_logged);
 
-        /* Look through operands for reads and writes. */
-        is_phi = ins->info->opcode == MVM_SSA_PHI;
-        for (i = 0; i < ins->info->num_operands; i++) {
-            /* Reads need usage tracking. */
-            if ((is_phi && i > 0)
-                || (!is_phi && (ins->info->operands[i] & MVM_operand_rw_mask) == MVM_operand_read_reg)) {
-                MVMSpeshFacts *facts = &(g->facts[ins->operands[i].reg.orig][ins->operands[i].reg.i]);
-                MVM_spesh_usages_add(tc, g, facts, ins);
-            }
-
-            /* Writes need the writing instruction to be specified. */
-            if ((is_phi && i == 0)
-                || (!is_phi && (ins->info->operands[i] & MVM_operand_rw_mask) == MVM_operand_write_reg)) {
-                MVMSpeshFacts *facts = &(g->facts[ins->operands[i].reg.orig][ins->operands[i].reg.i]);
-                facts->writer    = ins;
-            }
-        }
-
         /* Look for ops that are fact-interesting. */
         switch (ins->info->opcode) {
-        case MVM_OP_inc_i:
-        case MVM_OP_inc_u:
-        case MVM_OP_dec_i:
-        case MVM_OP_dec_u: {
-            /* These all read as well as write a value, so bump usages. */
-            MVMSpeshOperand reader;
-            reader.reg.orig = ins->operands[0].reg.orig;
-            reader.reg.i = ins->operands[0].reg.i - 1;
-            MVM_spesh_usages_add_by_reg(tc, g, reader, ins);
-            break;
-        }
         case MVM_OP_set:
             copy_facts(tc, g,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
@@ -487,6 +454,13 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
                 ins->operands[1].reg.orig, ins->operands[1].reg.i);
             break;
         case MVM_OP_sp_fastcreate:
+        case MVM_OP_sp_fastbox_i:
+        case MVM_OP_sp_fastbox_bi:
+        case MVM_OP_sp_fastbox_i_ic:
+        case MVM_OP_sp_fastbox_bi_ic:
+        case MVM_OP_sp_add_I:
+        case MVM_OP_sp_sub_I:
+        case MVM_OP_sp_mul_I:
             create_facts_with_type(tc, g,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 ((MVMSTable *)g->spesh_slots[ins->operands[2].lit_i16])->WHAT);
@@ -521,6 +495,9 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 ins->operands[2].reg.orig, ins->operands[2].reg.i);
             break;
+        case MVM_OP_trunc_i16:
+            trunc_i16_facts(tc, g, ins);
+            break;
         case MVM_OP_bootint:
             object_facts(tc, g,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
@@ -561,6 +538,27 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 tc->instance->boot_types.BOOTHash);
             break;
+        case MVM_OP_hllboolfor: {
+            MVMSpeshOperand *name_op = &ins->operands[1];
+            MVMSpeshFacts *namefacts = &g->facts[name_op->reg.orig][name_op->reg.i];
+            if (namefacts->writer && namefacts->writer->info->opcode == MVM_OP_const_s) {
+                MVMString *hllname = MVM_cu_string(tc, g->sf->body.cu, ins->operands[1].lit_str_idx);
+                MVMHLLConfig *hll = MVM_hll_get_config_for(tc, hllname);
+                if (hll->true_value && hll->false_value && STABLE(hll->true_value)->WHAT == STABLE(hll->false_value)->WHAT)
+                    create_facts_with_type(tc, g,
+                        ins->operands[0].reg.orig, ins->operands[0].reg.i,
+                        STABLE(hll->true_value)->WHAT);
+            }
+            break;
+        }
+        case MVM_OP_hllbool: {
+            MVMHLLConfig *hll = g->sf->body.cu->body.hll_config;
+            if (hll->true_value && hll->false_value && STABLE(hll->true_value)->WHAT == STABLE(hll->false_value)->WHAT)
+                create_facts_with_type(tc, g,
+                    ins->operands[0].reg.orig, ins->operands[0].reg.i,
+                    STABLE(hll->true_value)->WHAT);
+            break;
+        }
         case MVM_OP_hllboxtype_i:
             object_facts(tc, g,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
@@ -701,12 +699,10 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 ins->operands[4].reg.orig, ins->operands[4].reg.i);
             break;
-        case MVM_OP_cas_o:
-        case MVM_OP_atomicload_o: {
-            MVMSpeshOperand result = ins->operands[0];
-            g->facts[result.reg.orig][result.reg.i].flags |= MVM_SPESH_FACT_DECONTED;
+        case MVM_OP_setdispatcher:
+        case MVM_OP_setdispatcherfor:
+            g->sets_dispatcher = 1;
             break;
-        }
         case MVM_OP_sp_guard:
         case MVM_OP_sp_guardconc:
         case MVM_OP_sp_guardtype:
@@ -745,8 +741,8 @@ static void tweak_block_handler_usage(MVMThreadContext *tc, MVMSpeshGraph *g) {
 /* Kicks off fact discovery from the top of the (dominator) tree. */
 void MVM_spesh_facts_discover(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshPlanned *p,
         MVMuint32 is_specialized) {
-    /* The facts pass sets up usage normal usage information. */
-    add_bb_facts(tc, g, g->entry, p);
+    /* Set up normal usage information. */
+    MVM_spesh_usages_create_usage(tc, g);
     tweak_block_handler_usage(tc, g);
 
     /* We do an initial dead instruction pass before then computing the deopt
@@ -759,4 +755,7 @@ void MVM_spesh_facts_discover(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshPl
         MVM_spesh_eliminate_dead_ins(tc, g);
         MVM_spesh_usages_create_deopt_usage(tc, g);
     }
+
+    /* Finally, collect facts. */
+    add_bb_facts(tc, g, g->entry, p);
 }
